#include "header\gba.h"
#include "header\bg.h"
#include "header\timer.h"
#include "header\dma.h"
#include "header\irq.h"
#include "header\sound.h"
#include "header\sprite.h"
#include "header\blend.h"

#include "header\font\BGFont.h"
//#include "header\font\BitMapFont.h"

#include "lib.h"
#include "syscall.h"
#include "resource.h"

//Mode0
//BG2 \
//BG3 ͕

//---------------------------------------------------------------------------
//
#define BG_CX					8
#define BG_CY					8
#define BG_SCREEN_CX			30
#define BG_SCREEN_CY			20
#define BG_WORLD_SCREEN_CX		32
#define BG_WORLD_SCREEN_CY		32
#define BG_WORLD_SCREEN_WCX		256
#define BG_WORLD_SCREEN_WCY		256
#define BG_WRITE_SY				(20-3)

//is
#define STATE_INIT				1
#define STATE_INPUT				2
#define STATE_WRITE				3
#define STATE_CALC				4

//̓L[̃EFCg
#define KEY_WAIT				2

//g[N`
#define EOS						257
#define INT						258

//͕̍ő吔
#define MAX_STR_BUFFER			255

//---------------------------------------------------------------------------
typedef struct {
	u8 Chr;
	u8 Map;
	u8 Col;
} ST_BG;

typedef struct {
	s16 x;
	s16 y;
	s16 sy;
	u16 cx;
	u16 cy;
	u16 sprNo;
	u16 sprTile;
	u16 sprSize;
	u16 sprForm;
} ST_WRITER;

typedef struct {
	u8  x;
	u8  y;
	u8  cx;
	u8  cy;
	s16 minX;		//ړ
	s16 maxX;
	s16 minY;
	s16 maxY;
	u16 sprNo;
	u16 sprTile;
	u16 sprSize;
	u16 sprForm;
	u8  keyNow;		//ړp
	u8  keyOld;		//L[s[hp
	u8  keyCnt;		//L[s[hp
} ST_CURSOR;

typedef struct {
	u16 type;
	s32 val;
} ST_TOKEN;

typedef struct {
	u8  str[MAX_STR_BUFFER];
	s16 readCnt;
	s16 writeCnt;
} ST_BUFFER;

//---------------------------------------------------------------------------
ST_BG     BG2;
ST_BG     BG3;
ST_CURSOR Cursor;
ST_WRITER Writer;
ST_TOKEN  Token;
ST_BUFFER Buffer;

u8  State;			//istO
u16 BGvofs;			//y
//---------------------------------------------------------------------------
void StateInit();
void StateInput();
void StateWrite();
void StateCalc();

void InitWS();
void InitBG();
void InitSprite();
void InitCursor();
void InitBuffer();

void DrawFrame();
void DrawCursor();
void DrawWriter();
void DrawAnswer(s32 ans);

void ProcKeyInput();
void ProcDispWrite();
void ProcScroll();
s32  ProcCalc();

bool RepeatKey(ST_CURSOR* p);
bool ChkMoveCursor(ST_CURSOR* p);
void MoveCursor(ST_CURSOR* p);

void BGLineFeed();

void     lexar();
ST_TOKEN expr();
ST_TOKEN expr2();
ST_TOKEN term();
ST_TOKEN term2();
ST_TOKEN factor();

bool IsMatch(ST_TOKEN t, u16 type);
bool IsNumber(u8 chr);
//---------------------------------------------------------------------------
int main(void)
{
	State = STATE_INIT;

	for(;;)
	{
		WaitForVsync();

		switch(State)
		{
			case STATE_INIT:  StateInit();   break;
			case STATE_INPUT: StateInput();  break;
			case STATE_WRITE: StateWrite();  break;
			case STATE_CALC:  StateCalc();   break;
			default:
				DebugMessage((u8*)"Err: Main State");
				break;
		}
	}

	return 0;
}
//---------------------------------------------------------------------------
void StateInit()
{
	KeyInit();

	InitWS();
	InitBG();
	InitSprite();
	InitCursor();
	InitBuffer();

	DrawFrame();
	DrawCursor();
	DrawWriter();

	SetMode(MODE_0 | BG2_ENABLE | BG3_ENABLE | OBJ_ENABLE | OBJ_MAP_2D);
	State = STATE_INPUT;
}
//---------------------------------------------------------------------------
void InitWS()
{
	//EFCgRg[
	REG_WSCNT = CST_ROM0_1ST_3WAIT | CST_ROM0_2ND_1WAIT;
}
//---------------------------------------------------------------------------
void InitBG()
{
	//BG2 \
	//BG3 ͕

	BG2.Chr = 1;
	BG2.Map = 0;
	BG2.Col = 0;

	BG3.Chr = 1;	//BG2Ɠ
	BG3.Map = 1;
	BG3.Col = 0;	//BG2Ɠ

	REG_BG2CNT = (BG_SIZEA_256_256     | BG_COLOR_16         |
				  BG_CHARBASE(BG2.Chr) | BG_MAPBASE(BG2.Map) | 0);

	REG_BG3CNT = (BG_SIZEA_256_256     | BG_COLOR_16         |
				  BG_CHARBASE(BG3.Chr) | BG_MAPBASE(BG3.Map) | 3);

	BGSetPal16  (BG3.Col, (u16*)&BGFont_palette);
	BGSetChrData(BG3.Chr, (u16*)&BGFont_data, BGFont_DATA_SIZE);
}
//---------------------------------------------------------------------------
void InitSprite()
{
	SpriteInit();
	SpriteSetPal16((u16*)&cursor_pal_bin);
	SpriteSetData ((u16*)&cursor_img_bin, CURSOR_IMG_SIZE);
}
//---------------------------------------------------------------------------
void InitCursor()
{
	//\p
	Writer.x       = 0;
	Writer.y       = 0;
	Writer.sy      = 0;				//͊Jnn_
	Writer.cx      = 8;
	Writer.cy      = 8;
	Writer.sprNo   = 1;
	Writer.sprTile = 0;
	Writer.sprSize = SP_SIZE_16;
	Writer.sprForm = SP_SQUARE;

	SpriteSetSize(Writer.sprNo, Writer.sprSize, Writer.sprForm, SP_COLOR_16);
	SpriteSetChr (Writer.sprNo, Writer.sprTile);

	//͗p
	Cursor.x       =  0;
	Cursor.y       = 18;
	Cursor.cx      =  8;
	Cursor.cy      =  8;
	Cursor.minX    =  0;			//ړ
	Cursor.maxX    = 19;
	Cursor.minY    = 18;
	Cursor.maxY    = 18;
	Cursor.sprNo   =  2;
	Cursor.sprTile =  0;
	Cursor.sprSize = SP_SIZE_16;
	Cursor.sprForm = SP_SQUARE;

	SpriteSetSize(Cursor.sprNo, Cursor.sprSize, Cursor.sprForm, SP_COLOR_16);
	SpriteSetChr (Cursor.sprNo, Cursor.sprTile);
}
//---------------------------------------------------------------------------
void DrawFrame()
{
	BGDrawCls(BG3.Map, ' ');

	BGDrawStr(BG2.Map, 0, 17, (u8*)"------------------------------");
	BGDrawStr(BG2.Map, 0, 18, (u8*)"0123456789 +-*/%() ;          ");
	BGDrawStr(BG2.Map, 0, 19, (u8*)"                              ");
}
//---------------------------------------------------------------------------
void DrawCursor()
{
	s16 x  = Cursor.x * Cursor.cx;
	s16 y  = Cursor.y * Cursor.cy;
	u16 no = Cursor.sprNo;

	SpriteMove(no, x, y);
}
//---------------------------------------------------------------------------
void DrawWriter()
{
	s16 x  = Writer.x * Writer.cx;
	s16 y  = Writer.y * Writer.cy;
	u16 no = Writer.sprNo;

	SpriteMove(no, x, y);
}
//---------------------------------------------------------------------------
void StateInput()
{
	ProcKeyInput();
}
//---------------------------------------------------------------------------
void StateWrite()
{
	ProcDispWrite();
}
//---------------------------------------------------------------------------
void ProcKeyInput()
{
	ST_CURSOR* p = &Cursor;
	p->keyNow    = KeyInput();

	//s[h
	if( RepeatKey(p) == TRUE )
	{
		return;
	}

	if( ChkMoveCursor(p) == TRUE )
	{
		MoveCursor(p);
		DrawCursor();
	}

	if(p->keyNow & KEY_A)
	{
		State = STATE_WRITE;
	}
}
//---------------------------------------------------------------------------
//L[s[h
bool RepeatKey(ST_CURSOR* p)
{
	if(p->keyNow == 0x00 || p->keyNow != p->keyOld)
	{
		p->keyOld = p->keyNow;
		p->keyCnt = 0;
		return TRUE;
	}

	if(p->keyCnt <= KEY_WAIT)
	{
		p->keyCnt++;
		return TRUE;
	}

	p->keyCnt = 0;
	return FALSE;
}
//---------------------------------------------------------------------------
//ړ͈ړ\`FbN܂
bool ChkMoveCursor(ST_CURSOR* p)
{
	s16 tx1 = p->x;
	s16 ty1 = p->y;

	if(p->keyNow & KEY_UP   ) ty1--;
	if(p->keyNow & KEY_DOWN ) ty1++;
	if(p->keyNow & KEY_LEFT ) tx1--;
	if(p->keyNow & KEY_RIGHT) tx1++;

	if(tx1 < p->minX || ty1 < p->minY) return FALSE;
	if(tx1 > p->maxX || ty1 > p->maxY) return FALSE;

	return TRUE;
}
//---------------------------------------------------------------------------
//J[\̈ړ
void MoveCursor(ST_CURSOR* p)
{
	if(p->keyNow & KEY_UP   ) p->y--;
	if(p->keyNow & KEY_DOWN ) p->y++;
	if(p->keyNow & KEY_LEFT ) p->x--;
	if(p->keyNow & KEY_RIGHT) p->x++;
}
//---------------------------------------------------------------------------
void ProcDispWrite()
{
	ST_WRITER* wp = &Writer;
	ST_CURSOR* cp = &Cursor;

	//J[\WLN^擾
	u16 chr       = BGGetChar(BG2.Map, cp->x, cp->y);

	//擾LN^`
	BGDrawChar(BG3.Map, wp->x, (wp->y + BGvofs/BG_CY) % BG_WORLD_SCREEN_CY, BG3.Col, chr);
	Buffer.str[Buffer.writeCnt++] = chr;

	//܂Ԃ
	BGLineFeed();

	//͂A؂L`FbN܂
	if(chr == ';')
	{
		Writer.x = BG_SCREEN_CX;	//s܂
		BGLineFeed();

		State = STATE_CALC;
	}
	else
	{
		State = STATE_INPUT;
	}
}
//---------------------------------------------------------------------------
void ProcScroll()
{
	BGvofs = (BGvofs + BG_CY) % BG_WORLD_SCREEN_WCY;

	//̍sɏꂽe폜܂
	BGDrawStr(BG3.Map, 0, (Writer.y + BGvofs/BG_CY) % BG_WORLD_SCREEN_CY, (u8*)"                              ");

//	REG_BG3HOFS = BGhofs;		//x
	REG_BG3VOFS = BGvofs;		//y
}
//---------------------------------------------------------------------------
void StateCalc()
{
	s32 ans = ProcCalc();
	DrawAnswer(ans);

	Writer.x = BG_SCREEN_CX;	//s܂
	BGLineFeed();

	State = STATE_INPUT;
}
//---------------------------------------------------------------------------
s32 ProcCalc()
{
	Buffer.readCnt  = 0;
	Buffer.writeCnt = 0;

	lexar();
	ST_TOKEN t = expr();

	return t.val;
}
//---------------------------------------------------------------------------
void lexar()
{
	u8 chr;
	for(;;)
	{
		chr = Buffer.str[Buffer.readCnt++];

		if(chr != ' ')
		{
			break;
		}
	}

	switch(chr)
	{
		case '+':
		case '-':
		case '*':
		case '/':
		case '(':
		case ')':
			Token.type = chr;
			return;

		case ';':
			Token.type = EOS;
			return;

		default:
			if( IsNumber(chr) == FALSE )
			{
				Token.type = EOS;
				return;
			}

			s32 val = 0;
			do
			{
				//ӂ`FbNĂȂł
				val = (val * 10) + (chr - '0');
				chr = Buffer.str[Buffer.readCnt++];

			} while(IsNumber(chr) == TRUE);

			Buffer.readCnt--;
			Token.type = INT;
			Token.val  = val;
			return;
	}
}
//---------------------------------------------------------------------------
bool IsNumber(u8 chr)
{
	return (chr >= '0' && chr <= '9') ? TRUE : FALSE;
}
//---------------------------------------------------------------------------
ST_TOKEN expr()
{
	ST_TOKEN t = term();

	lexar();
	switch(Token.type)
	{
		case '+':
		case '-':
			return expr2(t);
		default:
			Buffer.readCnt--;
			return t;
	}
}
//---------------------------------------------------------------------------
ST_TOKEN expr2(ST_TOKEN t)
{
	while(Token.type == '+' || Token.type == '-')
	{
		u16 type = Token.type;

		lexar();
		ST_TOKEN t2 = term();

		if( IsMatch(t2, INT) == TRUE )
		{
			if(type == '+') t.val += t2.val;
			else            t.val -= t2.val;

			lexar();
		}
	}

	Buffer.readCnt--;
	return t;
}
//---------------------------------------------------------------------------
ST_TOKEN term()
{
	ST_TOKEN t = factor();

	lexar();
	for(;;)
	{
		switch(Token.type)
		{
			case '*':
			case '/':
				return t = term2(t);
			default:
				Buffer.readCnt--;
				return t;
		}
	}
}
//---------------------------------------------------------------------------
ST_TOKEN term2(ST_TOKEN t)
{
	while(Token.type == '*' || Token.type == '/')
	{
		u16 type = Token.type;
		lexar();

		if( IsMatch(Token, INT) == TRUE )
		{
			if(type == '*') t.val *= Token.val;
			else            t.val /= Token.val;

			lexar();
		}
	}

	Buffer.readCnt--;
	return t;
}
//---------------------------------------------------------------------------
ST_TOKEN factor()
{
	ST_TOKEN t;

	switch(Token.type)
	{
		case INT: return Token;
		case EOS: return Token;

		case '(':
			lexar();
			t = expr();
			if( IsMatch(Token, ')') == TRUE )
			{
				lexar();
				return t;
			}
			return Token;

		default : return Token;
	}
}
//---------------------------------------------------------------------------
bool IsMatch(ST_TOKEN t, u16 type)
{
	return t.type == type ? TRUE : FALSE;
}
//---------------------------------------------------------------------------
void BGLineFeed()
{
	ST_WRITER* wp = &Writer;

	//x
	if(wp->x + 1 >= BG_SCREEN_CX)
	{
		wp->x  = 0;
		wp->y += 1;
	}else{
		wp->x += 1;
	}

	//y
	if(wp->y >= BG_WRITE_SY)
	{
		wp->y = BG_WRITE_SY - 1;
		ProcScroll();
	}

	DrawWriter();
}
//---------------------------------------------------------------------------
void DrawAnswer(s32 ans)
{
	char msg[40];
	bprintf(msg, ":%d", ans);
	BGDrawStr(BG3.Map, 0, (Writer.y + BGvofs/BG_CY) % BG_WORLD_SCREEN_CY, (u8*)msg);

	DrawWriter();
}
//---------------------------------------------------------------------------
void InitBuffer()
{
	s16 i;
	for(i=0; i<MAX_STR_BUFFER; i++)
	{
		Buffer.str[i] = ' ';
	}
	Buffer.readCnt  = 0;
	Buffer.writeCnt = 0;
}
